(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('immer'), require('react'), require('redux'), require('redux-thunk'), require('memoizerific')) :
  typeof define === 'function' && define.amd ? define(['exports', 'immer', 'react', 'redux', 'redux-thunk', 'memoizerific'], factory) :
  (global = global || self, factory(global.EasyPeasy = {}, global.immer, global.React, global.Redux, global.ReduxThunk, global.memoizerific));
}(this, function (exports, produce, React, redux, reduxThunk, memoizerific) { 'use strict';

  var produce__default = 'default' in produce ? produce['default'] : produce;
  var React__default = 'default' in React ? React['default'] : React;
  reduxThunk = reduxThunk && reduxThunk.hasOwnProperty('default') ? reduxThunk['default'] : reduxThunk;
  memoizerific = memoizerific && memoizerific.hasOwnProperty('default') ? memoizerific['default'] : memoizerific;

  //

  var shallowequal = function shallowEqual(objA, objB, compare, compareContext) {
    var ret = compare ? compare.call(compareContext, objA, objB) : void 0;

    if (ret !== void 0) {
      return !!ret;
    }

    if (objA === objB) {
      return true;
    }

    if (typeof objA !== "object" || !objA || typeof objB !== "object" || !objB) {
      return false;
    }

    var keysA = Object.keys(objA);
    var keysB = Object.keys(objB);

    if (keysA.length !== keysB.length) {
      return false;
    }

    var bHasOwnProperty = Object.prototype.hasOwnProperty.bind(objB);

    // Test for A's keys different from B.
    for (var idx = 0; idx < keysA.length; idx++) {
      var key = keysA[idx];

      if (!bHasOwnProperty(key)) {
        return false;
      }

      var valueA = objA[key];
      var valueB = objB[key];

      ret = compare ? compare.call(compareContext, valueA, valueB, key) : void 0;

      if (ret === false || (ret === void 0 && valueA !== valueB)) {
        return false;
      }
    }

    return true;
  };

  var StoreContext = React.createContext();

  var isStateObject = function isStateObject(x) {
    return x !== null && typeof x === 'object' && !Array.isArray(x);
  };
  var get = function get(path, target) {
    return path.reduce(function (acc, cur) {
      return isStateObject(acc) ? acc[cur] : undefined;
    }, target);
  };
  var set = function set(path, target, value) {
    path.reduce(function (acc, cur, idx) {
      if (idx + 1 === path.length) {
        acc[cur] = value;
      } else {
        acc[cur] = acc[cur] || {};
      }

      return acc[cur];
    }, target);
  };

  function useStore(mapState, dependencies) {
    if (dependencies === void 0) {
      dependencies = [];
    }

    var store = React.useContext(StoreContext);

    var _useState = React.useState(mapState(store.getState())),
        state = _useState[0],
        setState = _useState[1]; // As our effect only fires on mount and unmount it won't have the state
    // changes visible to it, therefore we use a mutable ref to track this.


    var stateRef = React.useRef(state); // Helps avoid firing of events when unsubscribed, i.e. unmounted

    var isActive = React.useRef(true);
    React.useEffect(function () {
      var calculateState = function calculateState() {
        var newState = mapState(store.getState());
        isActive.current = true;

        if (newState === stateRef.current || isStateObject(newState) && isStateObject(stateRef.current) && shallowequal(newState, stateRef.current)) {
          // Do nothing
          return;
        }

        stateRef.current = newState;

        if (isActive.current) {
          setState(stateRef.current);
        }
      };

      calculateState();
      var unsubscribe = store.subscribe(calculateState);
      return function () {
        unsubscribe();
        isActive.current = false;
      };
    }, dependencies);
    return state;
  }
  function useActions(mapActions) {
    var store = React.useContext(StoreContext);
    return mapActions(store.dispatch);
  }
  function useDispatch() {
    var store = React.useContext(StoreContext);
    return store.dispatch;
  }
  function createTypedHooks() {
    return {
      useActions: useActions,
      useDispatch: useDispatch,
      useStore: useStore
    };
  }

  function _extends() {
    _extends = Object.assign || function (target) {
      for (var i = 1; i < arguments.length; i++) {
        var source = arguments[i];

        for (var key in source) {
          if (Object.prototype.hasOwnProperty.call(source, key)) {
            target[key] = source[key];
          }
        }
      }

      return target;
    };

    return _extends.apply(this, arguments);
  }

  var actionNameSymbol = '__actionName__';
  var actionSymbol = '__action__';
  var listenSymbol = '__listen__';
  var metaSymbol = '__meta__';
  var reducerSymbol = '__reducer__';
  var selectDependenciesSymbol = '__selectDependencies__';
  var selectImpSymbol = '__selectImp__';
  var selectStateSymbol = '__selectState__';
  var selectSymbol = '__select__';
  var thunkSymbol = '__thunk__';

  var actionName = function actionName(action) {
    return action[actionNameSymbol];
  };
  var thunkStartName = function thunkStartName(action) {
    return action[actionNameSymbol] + "(started)";
  };
  var thunkCompleteName = function thunkCompleteName(action) {
    return action[actionNameSymbol] + "(completed)";
  };
  var action = function action(fn) {
    fn[actionSymbol] = true;
    return fn;
  };
  var listen = function listen(fn) {
    fn[listenSymbol] = true;
    return fn;
  };
  var thunk = function thunk(fn) {
    fn[thunkSymbol] = true;
    return fn;
  };
  var select = function select(fn, dependencies) {
    fn[selectSymbol] = true;
    fn[selectDependenciesSymbol] = dependencies;
    fn[selectStateSymbol] = {};
    return fn;
  };
  var reducer = function reducer(fn) {
    fn[reducerSymbol] = true;
    return fn;
  };

  var maxSelectFnMemoize = 100;

  var tick = function tick() {
    return new Promise(function (resolve) {
      return setTimeout(resolve);
    });
  };

  var errorToPayload = function errorToPayload(err) {
    if (err instanceof Error) {
      return {
        message: err.message,
        stack: err.stack
      };
    }

    if (typeof err === 'string') {
      return err;
    }

    return undefined;
  };

  function createStoreInternals(_ref) {
    var disableInternalSelectFnMemoize = _ref.disableInternalSelectFnMemoize,
        initialState = _ref.initialState,
        injections = _ref.injections,
        model = _ref.model,
        reducerEnhancer = _ref.reducerEnhancer,
        references = _ref.references;

    var wrapFnWithMemoize = function wrapFnWithMemoize(x) {
      return !disableInternalSelectFnMemoize && typeof x === 'function' ? memoizerific(maxSelectFnMemoize)(x) : x;
    };

    var defaultState = {};
    var actionThunks = {};
    var actionCreators = {};
    var actionCreatorDict = {};
    var customReducers = [];
    var selectorReducers = [];
    var listenDefinitions = [];
    var thunkListenersDict = {};
    var actionListenersDict = {};
    var actionReducersDict = {};

    var recursiveExtractDefsFromModel = function recursiveExtractDefsFromModel(current, parentPath) {
      return Object.keys(current).forEach(function (key) {
        var value = current[key];
        var path = [].concat(parentPath, [key]);
        var meta = {
          parent: parentPath,
          path: path
        };

        if (typeof value === 'function') {
          if (value[actionSymbol]) {
            var name = "@action." + path.join('.');
            value[actionNameSymbol] = name;
            value[metaSymbol] = meta; // Action Reducer

            var actionReducer = value;
            actionReducer[actionNameSymbol] = name;
            actionReducersDict[name] = actionReducer;

            var actionCreator = function actionCreator(payload) {
              var result = references.dispatch({
                type: actionReducer[actionNameSymbol],
                payload: payload
              });
              return result;
            };

            actionCreator[actionNameSymbol] = name;
            actionCreatorDict[name] = actionCreator;
            set(path, actionCreators, actionCreator);
          } else if (value[thunkSymbol]) {
            var _name = "@thunk." + path.join('.');

            value[actionNameSymbol] = _name; // Thunk Action

            var action = function action(payload) {
              return value(get(parentPath, actionCreators), payload, {
                dispatch: references.dispatch,
                getState: function getState() {
                  return get(parentPath, references.getState());
                },
                getStoreState: references.getState,
                injections: injections,
                meta: meta
              });
            };

            set(path, actionThunks, action); // Thunk Action Creator

            var _actionCreator = function _actionCreator(payload) {
              return tick().then(function () {
                return references.dispatch({
                  type: _name + "(started)",
                  payload: payload
                });
              }).then(function () {
                return references.dispatch(function () {
                  return action(payload);
                });
              }).then(function (result) {
                references.dispatch({
                  type: _name + "(completed)",
                  payload: payload
                });
                return result;
              }).catch(function (err) {
                references.dispatch({
                  type: _name + "(failed)",
                  payload: payload,
                  error: errorToPayload(err)
                });
                throw err;
              });
            };

            _actionCreator[actionNameSymbol] = _name;
            actionCreatorDict[_name] = _actionCreator;
            set(path, actionCreators, _actionCreator);
          } else if (value[selectSymbol]) {
            value[selectStateSymbol] = {
              parentPath: parentPath,
              key: key,
              executed: false
            };
            selectorReducers.push(value);
          } else if (value[reducerSymbol]) {
            customReducers.push({
              path: path,
              reducer: value
            });
          } else if (value[listenSymbol]) {
            listenDefinitions.push(value);
            value[metaSymbol] = meta;
          } else {
            // eslint-disable-next-line no-console
            console.warn("Easy Peasy: Found a function at path " + path.join('.') + " in your model. Version 2 required that you wrap your action functions with the \"action\" helper");
          }
        } else if (isStateObject(value) && Object.keys(value).length > 0) {
          set(path, defaultState, {});
          recursiveExtractDefsFromModel(value, path);
        } else {
          // State
          var initialParentRef = get(parentPath, initialState);

          if (initialParentRef && key in initialParentRef) {
            set(path, defaultState, initialParentRef[key]);
          } else {
            set(path, defaultState, value);
          }
        }
      });
    };

    recursiveExtractDefsFromModel(model, []);
    selectorReducers.forEach(function (selector) {
      selector[selectImpSymbol] = function (state) {
        return wrapFnWithMemoize(selector(state));
      };
    });
    listenDefinitions.forEach(function (def) {
      var on = function on(target, handler) {
        if (typeof handler !== 'function') {
          return;
        }

        var meta = def[metaSymbol];
        handler[metaSymbol] = meta;

        if (!handler[actionSymbol] && !handler[thunkSymbol]) {
          {
            // eslint-disable-next-line
            console.warn("Easy Peasy: you must provide either an \"action\" or \"thunk\" to your listeners. Found an invalid handler at \"" + meta.path.join('.') + "\"");
          }

          return;
        }

        var name;

        if (typeof target === 'function' && target[actionNameSymbol] && actionCreatorDict[target[actionNameSymbol]]) {
          if (target[thunkSymbol]) {
            name = thunkCompleteName(target);
          } else {
            name = target[actionNameSymbol];
          }
        } else if (typeof target === 'string') {
          name = target;
        }

        if (name) {
          if (handler[thunkSymbol]) {
            thunkListenersDict[name] = thunkListenersDict[name] || [];
            thunkListenersDict[name].push(handler);
          } else {
            actionListenersDict[name] = actionListenersDict[name] || [];
            actionListenersDict[name].push({
              path: meta.parent,
              handler: handler
            });
          }
        }
      };

      def(on);
    });

    var runSelectorReducer = function runSelectorReducer(state, selector) {
      var _selector$selectState = selector[selectStateSymbol],
          parentPath = _selector$selectState.parentPath,
          key = _selector$selectState.key,
          executed = _selector$selectState.executed;

      if (executed) {
        return state;
      }

      var dependencies = selector[selectDependenciesSymbol];
      var stateAfterDependencies = dependencies ? dependencies.reduce(runSelectorReducer, state) : state;
      var newState = stateAfterDependencies;

      if (parentPath.length > 0) {
        var target = get(parentPath, stateAfterDependencies);

        if (target) {
          if (!selector.prevState || selector.prevState !== state) {
            var newValue = selector[selectImpSymbol](target);
            newState = produce__default(state, function (draft) {
              var updateTarget = get(parentPath, draft);
              updateTarget[key] = newValue;
            });
            selector.prevState = newState;
          }
        }
      } else if (!selector.prevState || selector.prevState !== state) {
        var _newValue = selector[selectImpSymbol](stateAfterDependencies);

        newState = produce__default(state, function (draft) {
          draft[key] = _newValue;
        });
        selector.prevState = newState;
      }

      selector[selectStateSymbol].executed = true;
      return newState;
    };

    var runSelectors = function runSelectors(state) {
      return selectorReducers.reduce(runSelectorReducer, state);
    };

    var createReducer = function createReducer() {
      var runActionReducerAtPath = function runActionReducerAtPath(state, action, actionReducer, path) {
        var current = get(path, state);

        if (path.length === 0) {
          return produce__default(state, function (_draft) {
            return actionReducer(_draft, action.payload);
          });
        }

        return produce__default(state, function (draft) {
          set(actionReducer[metaSymbol].parent, draft, produce__default(current, function (_draft) {
            return actionReducer(_draft, action.payload);
          }));
        });
      };

      var reducerForActions = function reducerForActions(state, action) {
        var actionReducer = actionReducersDict[action.type];

        if (actionReducer) {
          return runActionReducerAtPath(state, action, actionReducer, actionReducer[metaSymbol].parent);
        }

        return state;
      };

      var reducerForListeners = function reducerForListeners(state, action) {
        var actionListeners = actionListenersDict[action.type];

        if (actionListeners) {
          return actionListeners.reduce(function (newState, _ref2) {
            var path = _ref2.path,
                handler = _ref2.handler;
            return runActionReducerAtPath(newState, action, handler, path);
          }, state);
        }

        return state;
      };

      var reducerForCustomReducers = function reducerForCustomReducers(state, action) {
        return produce__default(state, function (draft) {
          customReducers.forEach(function (_ref3) {
            var p = _ref3.path,
                red = _ref3.reducer;
            var current = get(p, draft);
            set(p, draft, red(current, action));
          });
        });
      };

      var isInitial = true;

      var selectorsReducer = function selectorsReducer(state) {
        var stateAfterSelectors = runSelectors(state);
        isInitial = false;
        selectorReducers.forEach(function (selector) {
          selector[selectStateSymbol].executed = false;
        });
        return stateAfterSelectors;
      };

      var rootReducer = function rootReducer(state, action) {
        var stateAfterActions = reducerForActions(state, action);
        var stateAfterListeners = reducerForListeners(stateAfterActions, action);
        var stateAfterCustomReducers = reducerForCustomReducers(stateAfterListeners, action);

        if (state !== stateAfterCustomReducers || isInitial) {
          return selectorsReducer(stateAfterCustomReducers);
        }

        return stateAfterCustomReducers;
      };

      return rootReducer;
    };

    return {
      reducer: reducerEnhancer(createReducer()),
      defaultState: defaultState,
      actionCreators: actionCreators,
      thunkListenersDict: thunkListenersDict
    };
  }

  function createStore(model, options) {
    if (options === void 0) {
      options = {};
    }

    var _options = options,
        compose = _options.compose,
        _options$devTools = _options.devTools,
        devTools = _options$devTools === void 0 ? true : _options$devTools,
        _options$disableInter = _options.disableInternalSelectFnMemoize,
        disableInternalSelectFnMemoize = _options$disableInter === void 0 ? false : _options$disableInter,
        _options$initialState = _options.initialState,
        initialState = _options$initialState === void 0 ? {} : _options$initialState,
        injections = _options.injections,
        _options$mockActions = _options.mockActions,
        mockActions = _options$mockActions === void 0 ? false : _options$mockActions,
        _options$middleware = _options.middleware,
        middleware = _options$middleware === void 0 ? [] : _options$middleware,
        _options$reducerEnhan = _options.reducerEnhancer,
        reducerEnhancer = _options$reducerEnhan === void 0 ? function (rootReducer) {
      return rootReducer;
    } : _options$reducerEnhan;

    var modelDefinition = _extends({}, model, {
      logFullState: thunk(function (actions, payload, _ref) {
        var getState = _ref.getState;
        // eslint-disable-next-line no-console
        console.log(JSON.stringify(getState(), null, 2));
      }),
      replaceState: action(function (state, payload) {
        return payload;
      })
    });

    var references = {};
    var mockedActions = [];

    var mockActionsMiddlware = function mockActionsMiddlware() {
      return function (next) {
        return function (action) {
          if (mockActions) {
            mockedActions.push(action);
            return undefined;
          }

          return next(action);
        };
      };
    };

    var dispatchThunkListeners = function dispatchThunkListeners(name, payload) {
      var listensForAction = references.internals.thunkListenersDict[name];
      return listensForAction && listensForAction.length > 0 ? Promise.all(listensForAction.map(function (listenForAction) {
        return listenForAction(get(listenForAction[metaSymbol].parent, references.internals.actionCreators), payload, {
          dispatch: references.dispatch,
          getState: function getState() {
            return get(listenForAction[metaSymbol].parent, references.getState());
          },
          getStoreState: references.getState,
          injections: injections,
          meta: listenForAction[metaSymbol]
        });
      })) : Promise.resolve();
    };

    var dispatchActionStringListeners = function dispatchActionStringListeners() {
      return function (next) {
        return function (action) {
          if (references.internals.thunkListenersDict[action.type]) {
            dispatchThunkListeners(action.type, action.payload);
          }

          return next(action);
        };
      };
    };

    var composeEnhancers = compose || (devTools && typeof window !== 'undefined' && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ : redux.compose);

    var bindStoreInternals = function bindStoreInternals(state) {
      references.internals = createStoreInternals({
        disableInternalSelectFnMemoize: disableInternalSelectFnMemoize,
        initialState: state,
        injections: injections,
        model: modelDefinition,
        reducerEnhancer: reducerEnhancer,
        references: references
      });
    };

    bindStoreInternals(initialState);
    var store = redux.createStore(references.internals.reducer, references.internals.defaultState, composeEnhancers(redux.applyMiddleware.apply(void 0, [reduxThunk, dispatchActionStringListeners, mockActionsMiddlware].concat(middleware))));

    store.getMockedActions = function () {
      return [].concat(mockedActions);
    };

    store.clearMockedActions = function () {
      mockedActions = [];
    };

    references.dispatch = store.dispatch;
    references.getState = store.getState; // attach the action creators to dispatch

    var bindActionCreators = function bindActionCreators(actionCreators) {
      Object.keys(store.dispatch).forEach(function (actionsKey) {
        delete store.dispatch[actionsKey];
      });
      Object.keys(actionCreators).forEach(function (key) {
        store.dispatch[key] = actionCreators[key];
      });
    };

    bindActionCreators(references.internals.actionCreators);

    var rebindStore = function rebindStore() {
      bindStoreInternals(store.getState());
      store.replaceReducer(references.internals.reducer);
      store.dispatch.replaceState(references.internals.defaultState);
      bindActionCreators(references.internals.actionCreators);
    };

    store.addModel = function (key, modelForKey) {
      if (modelDefinition[key] && "development" !== 'production') {
        // eslint-disable-next-line no-console
        console.warn("The store model already contains a model definition for \"" + key + "\"");
        store.removeModel(key);
      }

      modelDefinition[key] = modelForKey;
      rebindStore();
    };

    store.removeModel = function (key) {
      if (!modelDefinition[key]) {
        {
          // eslint-disable-next-line no-console
          console.warn("The store model does not contain a model definition for \"" + key + "\"");
        }

        return;
      }

      delete modelDefinition[key];
      rebindStore();
    };

    return store;
  }

  var StoreProvider = function StoreProvider(_ref) {
    var children = _ref.children,
        store = _ref.store;
    return React__default.createElement(StoreContext.Provider, {
      value: store
    }, children);
  };

  /**
   * immer is an implementation detail, so we are not going to use its auto freeze
   * behaviour, which throws errors if trying to mutate state. It's also risky
   * for production builds as has a perf overhead.
   *
   * @see https://github.com/mweststrate/immer#auto-freezing
   */

  produce.setAutoFreeze(false);

  exports.action = action;
  exports.actionName = actionName;
  exports.createStore = createStore;
  exports.createTypedHooks = createTypedHooks;
  exports.listen = listen;
  exports.reducer = reducer;
  exports.select = select;
  exports.StoreProvider = StoreProvider;
  exports.thunk = thunk;
  exports.thunkStartName = thunkStartName;
  exports.thunkCompleteName = thunkCompleteName;
  exports.useActions = useActions;
  exports.useDispatch = useDispatch;
  exports.useStore = useStore;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=easy-peasy.umd.development.js.map
